/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.internal;

import java.net.InetSocketAddress;
import java.nio.charset.Charset;
import java.time.Instant;

import org.apache.commons.lang3.builder.ToStringBuilder;
import org.apache.commons.lang3.builder.ToStringStyle;

import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Connection payload for the {@link JackieChannelHandler}. This class is used
 * by the channel handler to store the connection state and pass it on to type
 * adapters.
 * 
 * @author Sebastian Marsching
 */
public final class JackieConnectionPayload {

    /**
     * Creates a connection payload that represents the connected state.
     * 
     * @param charset
     *            character set to be used to encode or decode strings.
     * @param honorZeroPrecision
     *            defines how a numeric value is formatted if the associated
     *            precision is zero. If <code>true</code> a precision of zero
     *            means that no decimal digits shall be printed. If
     *            <code>false</code>, a precision of zero means that a default
     *            number of decimal digits shall be printed.
     * @param longString
     *            <code>true</code> if a {@code DBR_CHAR} should be treated as a
     *            string, <code><code>false</code> if it should be treated as an
     *            array of bytes.
     * @param nativeCount
     *            number of elements that the channel reports as its native
     *            count. Typically, this is the max. number of elements that may
     *            be returned by a read operation on the channel.
     * @param nativeDataType
     *            native data-type of the channel. This is the data-type that is
     *            internally used on the server-side. If a read or write
     *            operation happens with this type, usually no conversion is
     *            required.
     * @param readable
     *            <code>true</code> if the channel is readable,
     *            <code>false</code> if it is write-only or no access is allowed
     *            at all.
     * @param serverAddress
     *            socket address of the server to which the channel is connected
     *            (may be <code>null</code> if this information is not
     *            available).
     * @param writable
     *            <code>true</code> if the channel is writable,
     *            <code>false</code> if it is read-only or no access is allowed
     *            at all.
     * @return payload representing the specified connection state.
     */
    public static JackieConnectionPayload connected(Charset charset,
            boolean honorZeroPrecision, boolean longString, int nativeCount,
            ChannelAccessValueType nativeDataType, boolean readable,
            InetSocketAddress serverAddress, boolean writable) {
        // This method is only used internally, so we use assertions instead of
        // real checks.
        assert (nativeDataType != null);
        return new JackieConnectionPayload(charset, true, honorZeroPrecision,
                longString, nativeCount, nativeDataType, readable,
                serverAddress, writable, writable);
    }

    /**
     * Creates a connection payload that represents the disconnected state. A
     * payload for the disconnected state must only be created if there already
     * is a payload for the previous connection state (typically this payload
     * represents the connected state).
     * 
     * @param oldConnectionPayload
     *            payload of the previous connection state.
     * @return payload representing the disconnected state.
     */
    public static JackieConnectionPayload disconnected(
            JackieConnectionPayload oldConnectionPayload) {
        // It might seem strange to assume that we always have an old connection
        // payload when creating one for the disconnected state. However, we
        // only create a disconnected connection payload when we have been
        // connected before, so this assumption is valid.
        return new JackieConnectionPayload(oldConnectionPayload.getCharset(),
                false, oldConnectionPayload.isHonorZeroPrecision(),
                oldConnectionPayload.isLongString(),
                oldConnectionPayload.getNativeCount(),
                oldConnectionPayload.getNativeDataType(), false, null, false,
                oldConnectionPayload.isWritableWhenLastConnected());
    }

    private final Charset charset;
    private final Instant disconnectTime;
    private final boolean honorZeroPrecision;
    private final boolean longString;
    private final int nativeCount;
    private final ChannelAccessValueType nativeDataType;
    private final boolean readable;
    private final boolean writable;
    private final boolean writableWhenLastConnected;
    private InetSocketAddress serverAddress;

    private JackieConnectionPayload(Charset charset, boolean connected,
            boolean honorZeroPrecision, boolean longString, int nativeCount,
            ChannelAccessValueType nativeDataType, boolean readable,
            InetSocketAddress serverAddress, boolean writable,
            boolean writableWhenLastConnected) {
        // This constructor is only used internally, so we use an assertion
        // instead of a real check.
        assert (charset != null);
        assert (nativeDataType != null);
        this.charset = charset;
        if (connected) {
            this.disconnectTime = null;
        } else {
            this.disconnectTime = Instant.now();
        }
        this.honorZeroPrecision = honorZeroPrecision;
        this.longString = longString;
        this.nativeDataType = nativeDataType;
        this.nativeCount = nativeCount;
        this.readable = readable;
        this.serverAddress = serverAddress;
        this.writable = writable;
        this.writableWhenLastConnected = writableWhenLastConnected;
    }

    /**
     * Tells whether the channel is currently connected.
     * 
     * @return <code>true</code> if the channel is connected, <code>false</code>
     *         if it is disconnected.
     */
    public boolean isConnected() {
        return disconnectTime == null;
    }

    /**
     * Tells whether to use a precision of zero. This defines how a numeric
     * value is formatted if the associated precision is zero. If
     * <code>true</code> a precision of zero means that no decimal digits shall
     * be printed. If <code>false</code>, a precision of zero means that a
     * default number of decimal digits shall be printed.
     * 
     * @return <code>true</code> if a precision of zero should be honored,
     *         <code>false</code> if such a precision should be replaced with
     *         the default precision.
     */
    public boolean isHonorZeroPrecision() {
        return honorZeroPrecision;
    }

    /**
     * Returns the character set to be used to encode and decode strings. This
     * character set is only used for "long" strings. Regular strings are
     * already decoded by the client itself.
     * 
     * @return character set used to encode and decode strings.
     */
    public Charset getCharset() {
        return charset;
    }

    /**
     * Returns the time when the connection was interrupted or <code>null</code>
     * if the channel is currently connected.
     * 
     * @return time when the channel was disconnected or <code>null</code> if
     *         the channel is connected.
     */
    public Instant getDisconnectTime() {
        return disconnectTime;
    }

    /**
     * <p>
     * Returns the number of elements that the channel reports as its native
     * count. Typically, this is the max. number of elements that may be
     * returned by a read operation on the channel.
     * </p>
     * 
     * <p>
     * If the channel is currently disconnected, this is the count that was
     * reported the last time it was connected.
     * </p>
     * 
     * @return native element count of the channel.
     */
    public int getNativeCount() {
        return nativeCount;
    }

    /**
     * <p>
     * Returns the native data-type of the channel. This is the data-type that
     * is internally used on the server-side. If a read or write operation
     * happens with this type, usually no conversion is required.
     * </p>
     * 
     * <p>
     * If the channel is currently disconnected, this is the data type that was
     * reported the last time it was connected.
     * </p>
     * 
     * @return
     */
    public ChannelAccessValueType getNativeDataType() {
        return nativeDataType;
    }

    /**
     * Returns the socket address of the server to which the channel is
     * connected. Returns <code>null</code> if the channel is not connected or
     * this information is not available.
     * 
     * @return socket address of the Channel Access server or <code>null</code>.
     */
    public InetSocketAddress getServerAddress() {
        return serverAddress;
    }

    /**
     * Tells whether an array of bytes (type {@code DBR_CHAR}) should be treated
     * as a null-terminated string.
     * 
     * @return <code>true</code> if a {@code DBR_CHAR} should be treated as a
     *         long string, <code>false</code> if it should be treated as an
     *         array of bytes.
     */
    public boolean isLongString() {
        return longString;
    }

    /**
     * Tells whether the channel is readable.
     * 
     * @return <code>true</code> if it is readable (and connected),
     *         <code>false</code> otherwise.
     */
    public boolean isReadable() {
        return readable;
    }

    /**
     * Tells whether the channel is writable.
     * 
     * @return <code>true</code> if it is writable (and connected),
     *         <code>false</code> otherwise.
     */
    public boolean isWritable() {
        return writable;
    }

    /**
     * Tells whether the channel was writable the last time it was connected. If
     * the channel is currently connected, this is the same as reported by
     * {@link #isWritable()}.
     * 
     * @return <code>true</code> if the channel was writable the last time it
     *         was connected, <code>false</code> otherwise.
     */
    public boolean isWritableWhenLastConnected() {
        return writableWhenLastConnected;
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this, ToStringStyle.SHORT_PREFIX_STYLE)
                .append("charset", charset)
                .append("disconnectTime", disconnectTime)
                .append("honorZeroPrecision", honorZeroPrecision)
                .append("longString", longString)
                .append("nativeCount", nativeCount)
                .append("nativeDataType", nativeDataType)
                .append("readable", readable).append("writable", writable)
                .append("writableWhenLastConnected", writableWhenLastConnected)
                .append("serverAddress", serverAddress).build();
    }

}
